/****************************************************************************************/
/*  SOZ.H                                                                               */
/*                                                                                      */
/*  Author:                                                                             */
/*  Description:                                                                        */
/*                                                                                      */
/*  The contents of this file are subject to the Jet3D Public License                   */
/*  Version 1.02 (the "License"); you may not use this file except in                   */
/*  compliance with the License. You may obtain a copy of the License at                */
/*  http://www.jet3d.com                                                                */
/*                                                                                      */
/*  Software distributed under the License is distributed on an "AS IS"                 */
/*  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See                */
/*  the License for the specific language governing rights and limitations              */
/*  under the License.                                                                  */
/*                                                                                      */
/*  The Original Code is Jet3D, released December 12, 1999.                             */
/*  Copyright (C) 1996-1999 Eclipse Entertainment, L.L.C. All Rights Reserved           */
/*                                                                                      */
/****************************************************************************************/
#ifndef __COMPUTIL_SOZ_H
#define __COMPUTIL_SOZ_H

/*
 * soz : s-context-o-zero for coding small-alphabet contexts
 *
 *  uses a linear array, not a binary tree
 *
 */

#include "arithc.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef struct soz soz;

extern soz * sozCreate(arithInfo * arithinfo,int size,int totmax,int inc);
extern void sozEncode(soz *sc, int symbol);
extern int sozDecode(soz *sc);			
extern void sozHalve(soz *sc);
extern void sozFree(soz *sc);
extern void sozReset(soz *sc);

#ifdef __cplusplus
}
#endif

#endif

